@@ -44,14 +44,6 @@ parse_ms_long_to_monotonic_t(PyObject *val) {
     return ms_to_monotonic_t(PyLong_AsUnsignedLong(val));
 }
 
-static inline int
-resolve_mods(int kitty_mod, int mods) {
-    if (mods & GLFW_MOD_KITTY) {
-        mods = (mods & ~GLFW_MOD_KITTY) | kitty_mod;
-    }
-    return mods;
-}
-
 static WindowTitleIn
 window_title_in(PyObject *title_in) {
     const char *in = PyUnicode_AsUTF8(title_in);
@@ -115,6 +107,24 @@ window_logo_path(PyObject *src, Options *opts) { STR_SETTER(default_window_logo)
 
 #undef STR_SETTER
 
+static void
+parse_font_mod_size(PyObject *val, float *sz, AdjustmentUnit *unit) {
+    PyObject *mv = PyObject_GetAttrString(val, "mod_value");
+    if (mv) {
+        *sz = PyFloat_AsFloat(PyTuple_GET_ITEM(mv, 0));
+        long u = PyLong_AsLong(PyTuple_GET_ITEM(mv, 1));
+        switch (u) { case POINT: case PERCENT: case PIXEL: *unit = u; break; }
+    }
+}
+
+static void
+modify_font(PyObject *mf, Options *opts) {
+#define S(which) { PyObject *v = PyDict_GetItemString(mf, #which); if (v) parse_font_mod_size(v, &opts->which.val, &opts->which.unit); }
+    S(underline_position); S(underline_thickness); S(strikethrough_thickness); S(strikethrough_position);
+    S(cell_height); S(cell_width); S(baseline);
+#undef S
+}
+
 static MouseShape
 pointer_shape(PyObject *shape_name) {
     const char *name = PyUnicode_AsUTF8(shape_name);
@@ -207,21 +217,3 @@ tab_bar_margin_height(PyObject *val, Options *opts) {
     opts->tab_bar_margin_height.outer = PyFloat_AsDouble(PyTuple_GET_ITEM(val, 0));
     opts->tab_bar_margin_height.inner = PyFloat_AsDouble(PyTuple_GET_ITEM(val, 1));
 }
-
-#define read_adjust(name) { \
-    if (PyFloat_Check(al)) { \
-        opts->name##_frac = (float)PyFloat_AsDouble(al); \
-        opts->name##_px = 0; \
-    } else { \
-        opts->name##_frac = 0; \
-        opts->name##_px = (int)PyLong_AsLong(al); \
-    } \
-}
-
-static void
-adjust_line_height(PyObject *al, Options *opts) { read_adjust(adjust_line_height); }
-static void
-adjust_column_width(PyObject *al, Options *opts) { read_adjust(adjust_column_width); }
-static void
-adjust_baseline(PyObject *al, Options *opts) { read_adjust(adjust_baseline); }
-#undef read_adjust
