@@ -1,16 +1,16 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
 from contextlib import suppress
 from typing import (
-    TYPE_CHECKING, Any, Callable, Dict, FrozenSet, Generator, Iterable, List,
+    TYPE_CHECKING, Any, Callable, Dict, FrozenSet, Iterable, Iterator, List,
     NoReturn, Optional, Tuple, Type, Union, cast
 )
 
 from kitty.cli import get_defaults_from_seq, parse_args, parse_option_spec
 from kitty.cli_stub import RCOptions as R
-from kitty.constants import appname, running_in_kitty
+from kitty.constants import appname, list_kitty_resources, running_in_kitty
+from kitty.types import AsyncResponse
 
 if TYPE_CHECKING:
     from kitty.boss import Boss as B
@@ -37,7 +37,7 @@ class MatchError(ValueError):
     hide_traceback = True
 
     def __init__(self, expression: str, target: str = 'windows'):
-        ValueError.__init__(self, 'No matching {} for expression: {}'.format(target, expression))
+        ValueError.__init__(self, f'No matching {target} for expression: {expression}')
 
 
 class OpacityError(ValueError):
@@ -65,9 +65,9 @@ class PayloadGetter:
 
 no_response = NoResponse()
 payload_get = object()
-ResponseType = Optional[Union[bool, str]]
-CmdReturnType = Union[Dict[str, Any], List, Tuple, str, int, float, bool]
-CmdGenerator = Generator[CmdReturnType, None, None]
+ResponseType = Union[bool, str, None, NoResponse, AsyncResponse]
+CmdReturnType = Union[Dict[str, Any], List[Any], Tuple[Any, ...], str, int, float, bool]
+CmdGenerator = Iterator[CmdReturnType]
 PayloadType = Optional[Union[CmdReturnType, CmdGenerator]]
 PayloadGetType = PayloadGetter
 ArgsType = List[str]
@@ -75,32 +75,91 @@ ArgsType = List[str]
 
 MATCH_WINDOW_OPTION = '''\
 --match -m
-The window to match. Match specifications are of the form:
-:italic:`field:regexp`. Where field can be one of: id, title, pid, cwd, cmdline, num, env.
-You can use the :italic:`ls` command to get a list of windows. Note that for
-numeric fields such as id, pid and num the expression is interpreted as a number,
-not a regular expression. The field num refers to the window position in the current tab,
-starting from zero and counting clockwise (this is the same as the order in which the
-windows are reported by the :italic:`ls` command). The window id of the current window
-is available as the KITTY_WINDOW_ID environment variable. When using the :italic:`env` field
-to match on environment variables you can specify only the environment variable name or a name
-and value, for example, :italic:`env:MY_ENV_VAR=2`
+The window to match. Match specifications are of the form: :italic:`field:query`.
+Where :italic:`field` can be one of: :code:`id`, :code:`title`, :code:`pid`, :code:`cwd`, :code:`cmdline`, :code:`num`,
+:code:`env`, :code:`state` and :code:`recent`.
+:italic:`query` is the expression to match. Expressions can be either a number or a regular expression, and can be
+:ref:`combined using Boolean operators <search_syntax>`.
+
+The special value :code:`all` matches all windows.
+
+For numeric fields: :code:`id`, :code:`pid`, :code:`num` and :code:`recent`, the expression is interpreted as
+a number, not a regular expression.
+
+The field :code:`num` refers to the window position in the current tab, starting from zero and counting clockwise (this
+is the same as the order in which the windows are reported by the :ref:`kitty @ ls <at-ls>` command).
+
+The window id of the current window is available as the :envvar:`KITTY_WINDOW_ID` environment variable.
+
+The field :code:`recent` refers to recently active windows in the currently active tab, with zero being the currently
+active window, one being the previously active window and so on.
+
+When using the :code:`env` field to match on environment variables, you can specify only the environment variable name
+or a name and value, for example, :code:`env:MY_ENV_VAR=2`.
+
+The field :code:`state` matches on the state of the window. Supported states are:
+:code:`active`, :code:`focused`, :code:`needs_attention`, :code:`parent_active` and :code:`parent_focused`.
+Active windows are the windows that are active in their parent tab. There is only one focused window and it is the
+window to which keyboard events are delivered.
+
+Note that you can use the :ref:`kitty @ ls <at-ls>` command to get a list of windows.
 '''
 MATCH_TAB_OPTION = '''\
 --match -m
-The tab to match. Match specifications are of the form:
-:italic:`field:regexp`. Where field can be one of:
-id, index, title, window_id, window_title, pid, cwd, env, cmdline.
-You can use the :italic:`ls` command to get a list of tabs. Note that for
-numeric fields such as id, index and pid the expression is interpreted as a number,
-not a regular expression. When using title or id, first a matching tab is
-looked for and if not found a matching window is looked for, and the tab
-for that window is used. You can also use window_id and window_title to match
-the tab that contains the window with the specified id or title. The index number
-is used to match the nth tab in the currently active OS window.
+The tab to match. Match specifications are of the form: :italic:`field:query`.
+Where :italic:`field` can be one of: :code:`id`, :code:`index`, :code:`title`, :code:`window_id`, :code:`window_title`,
+:code:`pid`, :code:`cwd`, :code:`cmdline` :code:`env`, :code:`state` and :code:`recent`.
+:italic:`query` is the expression to match. Expressions can be either a number or a regular expression, and can be
+:ref:`combined using Boolean operators <search_syntax>`.
+
+The special value :code:`all` matches all tabs.
+
+For numeric fields: :code:`id`, :code:`index`, :code:`window_id`, :code:`pid` and :code:`recent`, the
+expression is interpreted as a number, not a regular expression.
+
+When using :code:`title` or :code:`id`, first a matching tab is looked for, and if not found a matching window is looked
+for, and the tab for that window is used.
+
+You can also use :code:`window_id` and :code:`window_title` to match the tab that contains the window with the specified
+id or title.
+
+The :code:`index` number is used to match the nth tab in the currently active OS window.
+The :code:`recent` number matches recently active tabs in the currently active OS window, with zero being the currently
+active tab, one the previously active tab and so on.
+
+When using the :code:`env` field to match on environment variables, you can specify only the environment variable name
+or a name and value, for example, :code:`env:MY_ENV_VAR=2`. Tabs containing any window with the specified environment
+variables are matched.
+
+The field :code:`state` matches on the state of the tab. Supported states are:
+:code:`active`, :code:`focused`, :code:`needs_attention`, :code:`parent_active` and :code:`parent_focused`.
+Active tabs are the tabs that are active in their parent OS window. There is only one focused tab
+and it is the tab to which keyboard events are delivered.
+
+Note that you can use the :ref:`kitty @ ls <at-ls>` command to get a list of tabs.
 '''
 
 
+class ParsingOfArgsFailed(ValueError):
+    pass
+
+
+class AsyncResponder:
+
+    def __init__(self, payload_get: PayloadGetType, window: Optional[Window]) -> None:
+        self.async_id: str = payload_get('async_id', missing='')
+        self.peer_id: int = payload_get('peer_id', missing=0)
+        self.window_id: int = getattr(window, 'id', 0)
+
+    def send_data(self, data: Any) -> None:
+        from kitty.remote_control import send_response_to_client
+        send_response_to_client(data=data, peer_id=self.peer_id, window_id=self.window_id, async_id=self.async_id)
+
+    def send_error(self, error: str) -> None:
+        from kitty.remote_control import send_response_to_client
+        send_response_to_client(error=error, peer_id=self.peer_id, window_id=self.window_id, async_id=self.async_id)
+
+
 class RemoteCommand:
 
     name: str = ''
@@ -109,11 +168,13 @@ class RemoteCommand:
     argspec: str = '...'
     options_spec: Optional[str] = None
     no_response: bool = False
+    response_timeout: float = 10.  # seconds
     string_return_is_error: bool = False
     args_count: Optional[int] = None
     args_completion: Optional[Dict[str, Tuple[str, Union[Callable[[], Iterable[str]], Tuple[str, ...]]]]] = None
     defaults: Optional[Dict[str, Any]] = None
-    options_class: Type = RCOptions
+    is_asynchronous: bool = False
+    options_class: Type[RCOptions] = RCOptions
 
     def __init__(self) -> None:
         self.desc = self.desc or self.short_desc
@@ -183,23 +244,29 @@ class RemoteCommand:
                     windows += list(tab)
         return windows
 
+    def create_async_responder(self, payload_get: PayloadGetType, window: Optional[Window]) -> AsyncResponder:
+        return AsyncResponder(payload_get, window)
+
     def message_to_kitty(self, global_opts: RCOptions, opts: Any, args: ArgsType) -> PayloadType:
         raise NotImplementedError()
 
     def response_from_kitty(self, boss: 'Boss', window: Optional['Window'], payload_get: PayloadGetType) -> ResponseType:
         raise NotImplementedError()
 
+    def cancel_async_request(self, boss: 'Boss', window: Optional['Window'], payload_get: PayloadGetType) -> None:
+        pass
+
 
 def cli_params_for(command: RemoteCommand) -> Tuple[Callable[[], str], str, str, str]:
-    return (command.options_spec or '\n').format, command.argspec, command.desc, '{} @ {}'.format(appname, command.name)
+    return (command.options_spec or '\n').format, command.argspec, command.desc, f'{appname} @ {command.name}'
 
 
 def parse_subcommand_cli(command: RemoteCommand, args: ArgsType) -> Tuple[Any, ArgsType]:
     opts, items = parse_args(args[1:], *cli_params_for(command), result_class=command.options_class)
     if command.args_count is not None and command.args_count != len(items):
         if command.args_count == 0:
-            raise SystemExit('Unknown extra argument(s) supplied to {}'.format(command.name))
-        raise SystemExit('Must specify exactly {} argument(s) for {}'.format(command.args_count, command.name))
+            raise SystemExit(f'Unknown extra argument(s) supplied to {command.name}')
+        raise SystemExit(f'Must specify exactly {command.args_count} argument(s) for {command.name}')
     return opts, items
 
 
@@ -214,18 +281,14 @@ def command_for_name(cmd_name: str) -> RemoteCommand:
     try:
         m = import_module(f'kitty.rc.{cmd_name}')
     except ImportError:
-        raise KeyError(f'{cmd_name} is not a known kitty remote control command')
+        raise KeyError(f'Unknown kitty remote control command: {cmd_name}')
     return cast(RemoteCommand, getattr(m, cmd_name))
 
 
 def all_command_names() -> FrozenSet[str]:
-    try:
-        from importlib.resources import contents
-    except ImportError:
-        from importlib_resources import contents  # type:ignore
 
     def ok(name: str) -> bool:
         root, _, ext = name.rpartition('.')
         return bool(ext in ('py', 'pyc', 'pyo') and root and root not in ('base', '__init__'))
 
-    return frozenset({x.rpartition('.')[0] for x in filter(ok, contents('kitty.rc'))})
+    return frozenset({x.rpartition('.')[0] for x in filter(ok, list_kitty_resources('kitty.rc'))})
