@@ -20,6 +20,10 @@ from typing import (
 
 from .child import ProcessDesc
 from .cli_stub import CLIOptions
+from .clipboard import (
+    get_clipboard_string, get_primary_selection, set_clipboard_string,
+    set_primary_selection
+)
 from .config import build_ansi_color_table
 from .constants import (
     appname, clear_handled_signals, config_dir, is_macos, wakeup_io_loop
@@ -34,23 +38,25 @@ from .fast_data_types import (
     KeyEvent, Screen, add_timer, add_window, cell_size_for_window,
     click_mouse_cmd_output, click_mouse_url, compile_program,
     current_os_window, encode_key_for_tty, get_boss, get_click_interval,
-    get_clipboard_string, get_options, init_cell_program, mark_os_window_dirty,
-    mouse_selection, move_cursor_to_mouse_if_in_prompt, pt_to_px,
-    set_clipboard_string, set_titlebar_color, set_window_logo,
-    set_window_padding, set_window_render_data, update_ime_position_for_window,
-    update_window_title, update_window_visibility, wakeup_main_loop
+    get_options, init_cell_program, mark_os_window_dirty, mouse_selection,
+    move_cursor_to_mouse_if_in_prompt, pt_to_px, set_titlebar_color,
+    set_window_logo, set_window_padding, set_window_render_data,
+    update_ime_position_for_window, update_window_title,
+    update_window_visibility, wakeup_main_loop
 )
 from .keys import keyboard_mode_name, mod_mask
-from .notify import NotificationCommand, handle_notification_cmd, sanitize_identifier_pat
+from .notify import (
+    NotificationCommand, handle_notification_cmd, sanitize_identifier_pat
+)
 from .options.types import Options
 from .rgb import to_color
 from .terminfo import get_capabilities
 from .types import MouseEvent, OverlayType, WindowGeometry, ac, run_once
 from .typing import BossType, ChildType, EdgeLiteral, TabType, TypedDict
 from .utils import (
-    docs_url, get_primary_selection, kitty_ansi_sanitizer_pat, load_shaders,
-    log_error, open_cmd, open_url, parse_color_set, path_from_osc7_url,
-    resolve_custom_file, resolved_shell, sanitize_title, set_primary_selection
+    docs_url, kitty_ansi_sanitizer_pat, load_shaders, log_error, open_cmd,
+    open_url, parse_color_set, path_from_osc7_url, resolve_custom_file,
+    resolved_shell, sanitize_title
 )
 
 MatchPatternType = Union[Pattern[str], Tuple[Pattern[str], Optional[Pattern[str]]]]
